"""Loader Factory, Fast Collate, CUDA Prefetcher

Prefetcher and Fast Collate inspired by NVIDIA APEX example at
https://github.com/NVIDIA/apex/commit/d5e2bb4bdeedd27b1dfaf5bb2b24d6c000dee9be#diff-cf86c282ff7fba81fad27a559379d5bf

Hacked together by / Copyright 2019, Ross Wightman
"""

from functools import partial
from typing import Callable, Optional, Tuple, Union

import torch
import torch.utils.data
from timm.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.data.dataset import ImageDataset, IterableImageDataset
from timm.data.distributed_sampler import OrderedDistributedSampler, RepeatAugSampler
from timm.data.loader import MultiEpochsDataLoader, PrefetchLoader, _worker_init, fast_collate
from torch.utils.data.dataloader import default_collate
from torch.utils.data.distributed import DistributedSampler

from .transforms_factory import create_transform


def create_loader(
    dataset: Union[ImageDataset, IterableImageDataset],
    input_size: Union[int, Tuple[int, int], Tuple[int, int, int]],
    batch_size: int,
    is_training: bool = False,
    no_aug: bool = False,
    re_prob: float = 0.0,
    re_mode: str = "const",
    re_count: int = 1,
    re_split: bool = False,
    train_crop_mode: Optional[str] = None,
    scale: Optional[Tuple[float, float]] = None,
    ratio: Optional[Tuple[float, float]] = None,
    hflip: float = 0.5,
    vflip: float = 0.0,
    color_jitter: float = 0.4,
    color_jitter_prob: Optional[float] = None,
    grayscale_prob: float = 0.0,
    gaussian_blur_prob: float = 0.0,
    auto_augment: Optional[str] = None,
    num_aug_repeats: int = 0,
    num_aug_splits: int = 0,
    trivial_augment: bool = False,
    rand_augment: bool = False,
    ra_num_ops: int = 2,
    ra_magnitude: int = 8,
    ra_magnitude_bins: int = 31,
    interpolation: str = "bilinear",
    mean: Tuple[float, ...] = IMAGENET_DEFAULT_MEAN,
    std: Tuple[float, ...] = IMAGENET_DEFAULT_STD,
    num_workers: int = 1,
    distributed: bool = False,
    crop_pct: Optional[float] = None,
    crop_mode: Optional[str] = None,
    crop_border_pixels: Optional[int] = None,
    collate_fn: Optional[Callable] = None,
    pin_memory: bool = False,
    fp16: bool = False,  # deprecated, use img_dtype
    img_dtype: torch.dtype = torch.float32,
    device: torch.device = torch.device("cuda"),
    use_prefetcher: bool = True,
    use_multi_epochs_loader: bool = False,
    persistent_workers: bool = True,
    worker_seeding: str = "all",
    tf_preprocessing: bool = False,
):
    """

    Args:
        dataset: The image dataset to load.
        input_size: Target input size (channels, height, width) tuple or size scalar.
        batch_size: Number of samples in a batch.
        is_training: Return training (random) transforms.
        no_aug: Disable augmentation for training (useful for debug).
        re_prob: Random erasing probability.
        re_mode: Random erasing fill mode.
        re_count: Number of random erasing regions.
        re_split: Control split of random erasing across batch size.
        scale: Random resize scale range (crop area, < 1.0 => zoom in).
        ratio: Random aspect ratio range (crop ratio for RRC, ratio adjustment factor for RKR).
        hflip: Horizontal flip probability.
        vflip: Vertical flip probability.
        color_jitter: Random color jitter component factors (brightness, contrast, saturation, hue).
            Scalar is applied as (scalar,) * 3 (no hue).
        color_jitter_prob: Apply color jitter with this probability if not None (for SimlCLR-like aug
        grayscale_prob: Probability of converting image to grayscale (for SimCLR-like aug).
        gaussian_blur_prob: Probability of applying gaussian blur (for SimCLR-like aug).
        auto_augment: Auto augment configuration string (see auto_augment.py).
        num_aug_repeats: Enable special sampler to repeat same augmentation across distributed GPUs.
        num_aug_splits: Enable mode where augmentations can be split across the batch.
        interpolation: Image interpolation mode.
        mean: Image normalization mean.
        std: Image normalization standard deviation.
        num_workers: Num worker processes per DataLoader.
        distributed: Enable dataloading for distributed training.
        crop_pct: Inference crop percentage (output size / resize size).
        crop_mode: Inference crop mode. One of ['squash', 'border', 'center']. Defaults to 'center' when None.
        crop_border_pixels: Inference crop border of specified # pixels around edge of original image.
        collate_fn: Override default collate_fn.
        pin_memory: Pin memory for device transfer.
        fp16: Deprecated argument for half-precision input dtype. Use img_dtype.
        img_dtype: Data type for input image.
        device: Device to transfer inputs and targets to.
        use_prefetcher: Use efficient pre-fetcher to load samples onto device.
        use_multi_epochs_loader:
        persistent_workers: Enable persistent worker processes.
        worker_seeding: Control worker random seeding at init.
        tf_preprocessing: Use TF 1.0 inference preprocessing for testing model ports.

    Returns:
        DataLoader
    """
    re_num_splits = 0
    if re_split:
        # apply RE to second half of batch if no aug split otherwise line up with aug split
        re_num_splits = num_aug_splits or 2
    dataset.transform = create_transform(
        input_size,
        is_training=is_training,
        no_aug=no_aug,
        train_crop_mode=train_crop_mode,
        scale=scale,
        ratio=ratio,
        hflip=hflip,
        vflip=vflip,
        color_jitter=color_jitter,
        color_jitter_prob=color_jitter_prob,
        grayscale_prob=grayscale_prob,
        gaussian_blur_prob=gaussian_blur_prob,
        auto_augment=auto_augment,
        trivial_augment=trivial_augment,
        rand_augment=rand_augment,
        ra_num_ops=ra_num_ops,
        ra_magnitude=ra_magnitude,
        ra_magnitude_bins=ra_magnitude_bins,
        interpolation=interpolation,
        mean=mean,
        std=std,
        crop_pct=crop_pct,
        crop_mode=crop_mode,
        crop_border_pixels=crop_border_pixels,
        re_prob=re_prob,
        re_mode=re_mode,
        re_count=re_count,
        re_num_splits=re_num_splits,
        tf_preprocessing=tf_preprocessing,
        use_prefetcher=use_prefetcher,
        separate=num_aug_splits > 0,
    )

    if isinstance(dataset, IterableImageDataset):
        # give Iterable datasets early knowledge of num_workers so that sample estimates
        # are correct before worker processes are launched
        dataset.set_loader_cfg(num_workers=num_workers)

    sampler = None
    if distributed and not isinstance(dataset, torch.utils.data.IterableDataset):
        if is_training:
            if num_aug_repeats:
                sampler = RepeatAugSampler(dataset, num_repeats=num_aug_repeats)
            else:
                sampler = DistributedSampler(dataset)
        else:
            # This will add extra duplicate entries to result in equal num
            # of samples per-process, will slightly alter validation results
            sampler = OrderedDistributedSampler(dataset)
    else:
        assert (
            num_aug_repeats == 0
        ), "RepeatAugment not currently supported in non-distributed or IterableDataset use"

    if collate_fn is None:
        collate_fn = fast_collate if use_prefetcher else default_collate

    loader_class = torch.utils.data.DataLoader
    if use_multi_epochs_loader:
        loader_class = MultiEpochsDataLoader

    loader_args = dict(
        batch_size=batch_size,
        shuffle=not isinstance(dataset, torch.utils.data.IterableDataset)
        and sampler is None
        and is_training,
        num_workers=num_workers,
        sampler=sampler,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
        drop_last=is_training,
        worker_init_fn=partial(_worker_init, worker_seeding=worker_seeding),
        persistent_workers=persistent_workers,
    )
    try:
        loader = loader_class(dataset, **loader_args)  # type: ignore[arg-type]
    except TypeError:
        loader_args.pop("persistent_workers")  # only in Pytorch 1.7+
        loader = loader_class(dataset, **loader_args)  # type: ignore[arg-type]
    if use_prefetcher:
        prefetch_re_prob = re_prob if is_training and not no_aug else 0.0
        loader = PrefetchLoader(
            loader,
            mean=mean,
            std=std,
            channels=input_size[0],  # type: ignore[arg-type]
            device=device,
            fp16=fp16,  # deprecated, use img_dtype
            img_dtype=img_dtype,
            re_prob=prefetch_re_prob,
            re_mode=re_mode,
            re_count=re_count,
            re_num_splits=re_num_splits,
        )

    return loader



"""
Modified from https://github.com/thuml/Transfer-Learning-Library/blob/master/tllib/regularization/co_tuning.py
"""

import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import tqdm
from timm import utils


class Relationship:
    """Learns the category relationship p(y_s|y_t) between source dataset and target dataset.

    Args:
        data_loader (torch.utils.data.DataLoader): A data loader of target dataset.
        classifier (torch.nn.Module): A classifier for Co-Tuning.
        device (torch.nn.Module): The device to run classifier.
        cache (str, optional): Path to find and save the relationship file.

    """

    def __init__(self, data_loader, classifier, device, args, cache=None):
        super().__init__()
        self.data_loader = data_loader
        self.classifier = classifier
        self.args = args
        self.device = device
        if cache is None or not os.path.exists(cache):
            source_predictions, target_labels = self.collect_labels()
            self.relationship = self.get_category_relationship(source_predictions, target_labels)
            if cache is not None and utils.is_primary(args):
                np.save(cache, self.relationship)
        else:
            self.relationship = np.load(cache)

    def __getitem__(self, category):
        return self.relationship[category]

    def collect_labels(self):
        """
        Collects predictions of target dataset by source model and corresponding ground truth class labels.

        Returns:
            - source_probabilities, [N, N_p], where N_p is the number of classes in source dataset
            - target_labels, [N], where 0 <= each number < N_t, and N_t is the number of classes in target dataset
        """

        print("Collecting labels to calculate relationship")
        source_predictions = []
        target_labels = []

        self.classifier.to(self.device)
        self.classifier.eval()
        with torch.no_grad():
            for x, label in tqdm.tqdm(self.data_loader):
                x = x.to(self.device)
                feat = self.classifier.forward_features(x)
                y_s = self.classifier.forward_head(feat, pre_logits=True)

                if self.args.distributed:
                    gather_tensors = [torch.zeros_like(y_s) for _ in range(self.args.world_size)]
                    torch.dist.all_gather(gather_tensors, y_s)
                    y_s = torch.cat(gather_tensors, dim=0)
                y_s = (
                    y_s - y_s.max(dim=1, keepdim=True).values
                )  # original implementation: https://github.com/thuml/CoTuning/blob/fc3eb54c9b44251c7cba557f9b90bdee5eca6ec3/module/relationship_learning.py#L41
                source_predictions.append(F.softmax(y_s, dim=1).detach().cpu().numpy())
                target_labels.append(label.cpu().numpy())

        return np.concatenate(source_predictions, 0), np.concatenate(target_labels, 0)

    def get_category_relationship(self, source_probabilities, target_labels):
        """
        The direct approach of learning category relationship p(y_s | y_t).

        Args:
            source_probabilities (numpy.array): [N, N_p], where N_p is the number of classes in source dataset
            target_labels (numpy.array): [N], where 0 <= each number < N_t, and N_t is the number of classes in target dataset

        Returns:
            Conditional probability, [N_c, N_p] matrix representing the conditional probability p(pre-trained class | target_class)
        """
        N_t = np.max(target_labels) + 1  # the number of target classes
        conditional = []
        for i in range(N_t):
            this_class = source_probabilities[target_labels == i]
            average = np.mean(this_class, axis=0, keepdims=True)
            conditional.append(average)
        return np.concatenate(conditional)


class CoTuningLoss(nn.Module):
    """
    The Co-Tuning loss in `Co-Tuning for Transfer Learning (NIPS 2020)
    <http://ise.thss.tsinghua.edu.cn/~mlong/doc/co-tuning-for-transfer-learning-nips20.pdf>`_.

    Inputs:
        - input: p(y_s) predicted by source classifier.
        - target: p(y_s|y_t), where y_t is the ground truth class label in target dataset.

    Shape:
        - input:  (b, N_p), where b is the batch size and N_p is the number of classes in source dataset
        - target: (b, N_p), where b is the batch size and N_p is the number of classes in source dataset
        - Outputs: scalar.
    """

    def __init__(self, relationship: Relationship, weight: float):
        super().__init__()
        self.relationship = relationship
        self.weight = weight

    def forward(self, feature, target, **kwargs):
        if target.ndim == 2:  # if using mixup
            target = target.argmax(dim=-1)
        y_s = self.relationship[target.cpu().numpy()]
        y_s_tensor = torch.from_numpy(y_s).to(feature.device).float()
        y_t = feature - feature.max(dim=1, keepdim=True).values
        y = -y_s_tensor * F.log_softmax(y_t, dim=-1)
        y = torch.mean(torch.sum(y, dim=-1))
        return self.weight * y



from .bss import BatchSpectralShrinkage
from .co_tuning import CoTuningLoss, Relationship
from .delta import BehavioralRegularization, L2Regularization, SPRegularization

__all__ = [
    "BatchSpectralShrinkage",
    "CoTuningLoss",
    "Relationship",
    "BehavioralRegularization",
    "L2Regularization",
    "SPRegularization",
]


"""
Modified from https://github.com/thuml/Transfer-Learning-Library/blob/master/tllib/regularization/delta.py
"""

import functools

import torch
import torch.nn as nn


class L2Regularization(nn.Module):
    r"""The L2 regularization of parameters :math:`w` can be described as:

    .. math::
        {\Omega} (w) = \dfrac{1}{2}  \Vert w\Vert_2^2 ,

    Args:
        model (torch.nn.Module):  The model to apply L2 penalty.

    Shape:
        - Output: scalar.
    """

    def __init__(self, model: nn.Module):
        super(L2Regularization, self).__init__()
        self.model = model

    def forward(self):
        output = 0.0
        for param in self.model.parameters():
            output += 0.5 * torch.norm(param) ** 2
        return output


class SPRegularization(nn.Module):
    r"""The SP (Starting Point) regularization from `Explicit inductive bias for transfer learning with convolutional networks
    (ICML 2018) <https://arxiv.org/abs/1802.01483>`_

    The SP regularization of parameters :math:`w` can be described as:

    .. math::
        {\Omega} (w) = \dfrac{1}{2}  \Vert w-w_0\Vert_2^2 ,

    where :math:`w_0` is the parameter vector of the model pretrained on the source problem, acting as the starting point (SP) in fine-tuning.


    Args:
        source_model (torch.nn.Module):  The source (starting point) model.
        target_model (torch.nn.Module):  The target (fine-tuning) model.

    Shape:
        - Output: scalar.
    """

    def __init__(self, source_model: nn.Module, target_model: nn.Module, weight: float = 0.5):
        super().__init__()
        assert weight > 0.0
        self.weight = weight
        self.target_model = target_model
        self.source_model = source_model

    def forward(self, **kwargs):
        output = 0.0
        source_weight = dict(self.source_model.named_parameters())
        for name, param in self.target_model.named_parameters():
            if param.requires_grad:
                output += self.weight * torch.norm(param - source_weight[name]) ** 2

        return output


class BehavioralRegularization(nn.Module):
    r"""The behavioral regularization from `DELTA:DEep Learning Transfer using Feature Map with Attention
    for convolutional networks (ICLR 2019) <https://openreview.net/pdf?id=rkgbwsAcYm>`_

    It can be described as:

    .. math::
        {\Omega} (w) = \sum_{j=1}^{N}   \Vert FM_j(w, \boldsymbol x)-FM_j(w^0, \boldsymbol x)\Vert_2^2 ,

    where :math:`w^0` is the parameter vector of the model pretrained on the source problem, acting as the starting point (SP) in fine-tuning,
    :math:`FM_j(w, \boldsymbol x)` is feature maps generated from the :math:`j`-th layer of the model parameterized with :math:`w`, given the input :math:`\boldsymbol x`.


    Inputs:
        layer_outputs_source (OrderedDict):  The dictionary for source model, where the keys are layer names and the values are feature maps correspondingly.

        layer_outputs_target (OrderedDict):  The dictionary for target model, where the keys are layer names and the values are feature maps correspondingly.

    Shape:
        - Output: scalar.

    """

    def __init__(self, source_model: nn.Module, weight: float = 0.01):
        super().__init__()
        self.source_model = source_model
        self.weight = weight

    def forward(self, x, feature, **kwargs):
        # output = 0.0
        # for fm_src, fm_tgt in zip(layer_outputs_source.values(), layer_outputs_target.values()):
        #     output += 0.5 * (torch.norm(fm_tgt - fm_src.detach()) ** 2)
        feat = self.source_model.forward_features(x).detach()
        pre_logits = self.source_model.forward_head(feat, pre_logits=True)
        output = self.weight * torch.norm(feature - pre_logits) ** 2
        return output


class AttentionBehavioralRegularization(nn.Module):
    r"""The behavioral regularization with attention from `DELTA:DEep Learning Transfer using Feature Map with Attention
    for convolutional networks (ICLR 2019) <https://openreview.net/pdf?id=rkgbwsAcYm>`_

    It can be described as:

    .. math::
        {\Omega} (w) = \sum_{j=1}^{N}  W_j(w) \Vert FM_j(w, \boldsymbol x)-FM_j(w^0, \boldsymbol x)\Vert_2^2 ,

    where
    :math:`w^0` is the parameter vector of the model pretrained on the source problem, acting as the starting point (SP) in fine-tuning.
    :math:`FM_j(w, \boldsymbol x)` is feature maps generated from the :math:`j`-th layer of the model parameterized with :math:`w`, given the input :math:`\boldsymbol x`.
    :math:`W_j(w)` is the channel attention of the :math:`j`-th layer of the model parameterized with :math:`w`.

    Args:
        channel_attention (list): The channel attentions of feature maps generated by each selected layer. For the layer with C channels, the channel attention is a tensor of shape [C].

    Inputs:
        layer_outputs_source (OrderedDict):  The dictionary for source model, where the keys are layer names and the values are feature maps correspondingly.

        layer_outputs_target (OrderedDict):  The dictionary for target model, where the keys are layer names and the values are feature maps correspondingly.

    Shape:
        - Output: scalar.

    """

    def __init__(self, channel_attention):
        super().__init__()
        self.channel_attention = channel_attention

    def forward(self, layer_outputs_source, layer_outputs_target):
        output = 0.0
        for i, (fm_src, fm_tgt) in enumerate(
            zip(layer_outputs_source.values(), layer_outputs_target.values())
        ):
            b, c, h, w = fm_src.shape
            fm_src = fm_src.reshape(b, c, h * w)
            fm_tgt = fm_tgt.reshape(b, c, h * w)

            distance = torch.norm(fm_tgt - fm_src.detach(), 2, 2)
            distance = c * torch.mul(self.channel_attention[i], distance**2) / (h * w)
            output += 0.5 * torch.sum(distance)

        return output


def get_attribute(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))



"""Checkpoint Saver

Track top-n training checkpoints and maintain recovery checkpoints on specified intervals.

Modified from timm.utils.checkpoint_saver.py.
"""

import glob
import operator
import os
import logging
import pickle
import torch

from timm.utils.model import unwrap_model, get_state_dict


_logger = logging.getLogger(__name__)


class CheckpointSaver:
    """
    Track top-n training checkpoints and maintain recovery checkpoints on specified intervals.

    Modified from timm.utils.checkpoint_saver.py.
    """

    def __init__(
        self,
        model,
        optimizer,
        args=None,
        model_ema=None,
        amp_scaler=None,
        checkpoint_prefix="checkpoint",
        recovery_prefix="recovery",
        checkpoint_dir="",
        recovery_dir="",
        decreasing=False,
        max_history=10,
        unwrap_fn=unwrap_model,
        resume=False,
    ):
        """
        Initialize CheckpointSaver.

        Args:
            model: Model to be saved
            optimizer: Optimizer to be saved
            args: Arguments to be saved
            model_ema: EMA model to be saved
            amp_scaler: AMP scaler to be saved
            checkpoint_prefix: Prefix for checkpoint filenames
            recovery_prefix: Prefix for recovery filenames
            checkpoint_dir: Directory for saving checkpoints
            recovery_dir: Directory for saving recoveries
            decreasing: Whether to use decreasing metric for sorting checkpoints
            max_history: Maximum number of checkpoints to keep
            unwrap_fn: Function to unwrap model before saving
            resume: Whether to restore from previous checkpoint saver state
        """
        # objects to save state_dicts of
        self.model = model
        self.optimizer = optimizer
        self.args = args
        self.model_ema = model_ema
        self.amp_scaler = amp_scaler

        # state
        self.checkpoint_files = []  # (filename, metric) tuples in order of decreasing betterness
        self.best_epoch = None
        self.best_metric = None
        self.curr_recovery_file = ""
        self.last_recovery_file = ""

        # config
        self.checkpoint_dir = checkpoint_dir
        self.recovery_dir = recovery_dir
        self.save_prefix = checkpoint_prefix
        self.recovery_prefix = recovery_prefix
        self.extension = ".pth.tar"
        self.decreasing = decreasing  # a lower metric is better if True
        self.cmp = operator.lt if decreasing else operator.gt
        self.max_history = max_history
        self.unwrap_fn = unwrap_fn
        assert self.max_history >= 1

        if resume:
            state_path = os.path.join(self.checkpoint_dir, "checkpoint_saver.pkl")
            try:
                with open(state_path, "rb") as f:
                    state_saver = pickle.load(f)
                self.__dict__.update(state_saver)
            except FileNotFoundError:
                _logger.warning(
                    "Resume requested but CheckpointSaver state-file not found. "
                    "This might lead to enexpected behavior. "
                    "Best model must not be the best model."
                )
            _logger.info("Restored CheckpointSaver state from '{}'".format(state_path))

    def save_checkpoint(self, epoch, metric=None):
        """
        Save checkpoint.

        Args:
            epoch: Epoch to be saved
            metric: Metric to be saved
        """
        assert epoch >= 0
        tmp_save_path = os.path.join(self.checkpoint_dir, "tmp" + self.extension)
        last_save_path = os.path.join(self.checkpoint_dir, "last" + self.extension)
        self._save(tmp_save_path, epoch, metric)
        if os.path.exists(last_save_path):
            os.unlink(last_save_path)  # required for Windows support.
        os.rename(tmp_save_path, last_save_path)
        worst_file = self.checkpoint_files[-1] if self.checkpoint_files else None
        if (
            len(self.checkpoint_files) < self.max_history
            or metric is None
            or self.cmp(metric, worst_file[1])  # type: ignore
        ):
            if len(self.checkpoint_files) >= self.max_history:
                self._cleanup_checkpoints(1)
            filename = "-".join([self.save_prefix, str(epoch)]) + self.extension
            save_path = os.path.join(self.checkpoint_dir, filename)
            os.link(last_save_path, save_path)
            self.checkpoint_files.append((save_path, metric))
            self.checkpoint_files = sorted(
                self.checkpoint_files, key=lambda x: x[1], reverse=not self.decreasing
            )  # sort in descending order if a lower metric is not better

            checkpoints_str = "Current checkpoints:\n"
            for c in self.checkpoint_files:
                checkpoints_str += " {}\n".format(c)
            _logger.info(checkpoints_str)

            if metric is not None and (
                self.best_metric is None or self.cmp(metric, self.best_metric)
            ):
                self.best_epoch = epoch
                self.best_metric = metric
                best_save_path = os.path.join(self.checkpoint_dir, "model_best" + self.extension)
                if os.path.exists(best_save_path):
                    os.unlink(best_save_path)
                os.link(last_save_path, best_save_path)
        # save checkpoint saver state
        self._save_self_state()

        return (None, None) if self.best_metric is None else (self.best_metric, self.best_epoch)

    def _save(self, save_path, epoch, metric=None):
        """
        Save checkpoint to file.

        Args:
            save_path: Path to save checkpoint
            epoch: Epoch to be saved
            metric: Metric to be saved
        """
        save_state = {
            "epoch": epoch,
            "arch": type(self.model).__name__.lower(),
            "state_dict": get_state_dict(self.model, self.unwrap_fn),
            "optimizer": self.optimizer.state_dict(),
            "version": 2,  # version < 2 increments epoch before save
        }
        if self.args is not None:
            save_state["arch"] = self.args.model
            save_state["args"] = self.args
        if self.amp_scaler is not None:
            save_state[self.amp_scaler.state_dict_key] = self.amp_scaler.state_dict()
        if self.model_ema is not None:
            save_state["state_dict_ema"] = get_state_dict(self.model_ema, self.unwrap_fn)
        if metric is not None:
            save_state["metric"] = metric
        torch.save(save_state, save_path)

    def _save_self_state(self):
        """
        Save CheckpointSaver state to file.
        """
        # save checkpoint saver state
        state_path = os.path.join(self.checkpoint_dir, "checkpoint_saver.pkl")
        state_saver = {
            "best_epoch": self.best_epoch,
            "best_metric": self.best_metric,
            "checkpoint_files": self.checkpoint_files,
            "curr_recovery_file": self.curr_recovery_file,
            "last_recovery_file": self.last_recovery_file,
        }
        with open(state_path, "wb") as f:
            pickle.dump(state_saver, f, protocol=pickle.HIGHEST_PROTOCOL)

    def _cleanup_checkpoints(self, trim=0):
        """
        Cleanup checkpoints.

        Args:
            trim: Number of checkpoints to keep
        """
        trim = min(len(self.checkpoint_files), trim)
        delete_index = self.max_history - trim
        if delete_index < 0 or len(self.checkpoint_files) <= delete_index:
            return
        to_delete = self.checkpoint_files[delete_index:]
        for d in to_delete:
            try:
                _logger.debug("Cleaning checkpoint: {}".format(d))
                os.remove(d[0])
            except Exception as e:
                _logger.error("Exception '{}' while deleting checkpoint".format(e))
        self.checkpoint_files = self.checkpoint_files[:delete_index]

    def save_recovery(self, epoch, batch_idx=0):
        """
        Save recovery checkpoint.

        Args:
            epoch: Epoch to be saved
            batch_idx: Batch index to be saved
        """
        assert epoch >= 0
        filename = "-".join([self.recovery_prefix, str(epoch), str(batch_idx)]) + self.extension
        save_path = os.path.join(self.recovery_dir, filename)
        self._save(save_path, epoch)
        if os.path.exists(self.last_recovery_file):
            try:
                _logger.debug("Cleaning recovery: {}".format(self.last_recovery_file))
                os.remove(self.last_recovery_file)
            except Exception as e:
                _logger.error("Exception '{}' while removing {}".format(e, self.last_recovery_file))
        self.last_recovery_file = self.curr_recovery_file
        self.curr_recovery_file = save_path

    def find_recovery(self):
        """
        Find the latest recovery checkpoint.

        Returns:
            Path to the latest recovery checkpoint
        """
        recovery_path = os.path.join(self.recovery_dir, self.recovery_prefix)
        files = glob.glob(recovery_path + "*" + self.extension)
        files = sorted(files)
        return files[0] if len(files) else ""



import torch.nn as nn


class FineTuningModel(nn.Module):
    """
    A wrapper of the original model for fine-tuning.
    """

    def __init__(self, model: nn.Module):
        """
        Args:
            model (nn.Module): The original model.
        """
        super().__init__()
        self._orig_mod = model

    def forward(self, x):
        """
        Forward a batch of input data through the model.

        Args:
            x (torch.Tensor): The input data.

        Returns:
            logits (torch.Tensor): The output of the model.
            pre_logits (torch.Tensor): The output of the model before the final
                classification layer.
        """
        # Compute the feature representation of the input data
        feat = self._orig_mod.forward_features(x)

        # Compute the output of the model before the final classification layer
        pre_logits = self._orig_mod.forward_head(feat, pre_logits=True)

        # Compute the output of the model
        logits = self._orig_mod.forward_head(feat)

        return logits, pre_logits


def reset_head(model, args):
    """
    Reset the head of a model to the given number of classes.

    Args:
        model (nn.Module): The model to reset the head of.
        args: The arguments containing the number of classes to reset the head to.

    Returns:
        The model with the head reset to the given number of classes.
    """
    pool = None
    if hasattr(model, "global_pool"):
        pool = getattr(model, "global_pool")
        if not isinstance(pool, str):
            pool = pool.pool_type
    model.reset_classifier(num_classes=args.num_classes, global_pool=pool)


def create_finetuning_model(
    model: nn.Module, args
) -> nn.Module:
    """
    Create a fine-tuning model from a given model.

    Args:
        model (nn.Module): The model to create a fine-tuning model from.
        args: Arguments-object for creating the fine-tuning model.

    Returns:
        nn.Module: The created fine-tuning model.
    """
    # reset the classifier/head of the model
    reset_head(model, args)
    
    total_params: int = sum(param.numel() for param in model.parameters())

    #freeze model parameters
    if args.linear_probing:
        for param in model.parameters():
            param.requires_grad = False
    else:
        num_frozen: int = 0
        for param in model.parameters():
            if num_frozen < args.pct_to_freeze * total_params:
                param.requires_grad = False
                num_frozen += param.numel()
            else:
                break
    # unfreezing classifier/head, always used for finetuning 
    for param in model.get_classifier().parameters():
        param.requires_grad = True

    trainable_params: int = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print(f"Total params: {total_params}, Trainable params: {trainable_params}")

    return FineTuningModel(model)



"""
@author: Yifei Ji
@contact: jiyf990330@163.com
"""

import torch
import torch.nn as nn


class BatchSpectralShrinkage(nn.Module):
    r"""The regularization term in `Catastrophic Forgetting Meets Negative Transfer:
    Batch Spectral Shrinkage for Safe Transfer Learning (NIPS 2019) <https://proceedings.neurips.cc/paper/2019/file/c6bff625bdb0393992c9d4db0c6bbe45-Paper.pdf>`_.

    The BSS regularization of feature matrix :math:`F` can be described as:

    .. math::
        L_{bss}(F) = \sum_{i=1}^{k} \sigma_{-i}^2 ,

    where :math:`k` is the number of singular values to be penalized, :math:`\sigma_{-i}` is the :math:`i`-th smallest singular value of feature matrix :math:`F`.

    All the singular values of feature matrix :math:`F` are computed by `SVD`:

    .. math::
        F = U\Sigma V^T,

    where the main diagonal elements of the singular value matrix :math:`\Sigma` is :math:`[\sigma_1, \sigma_2, ..., \sigma_b]`.


    Args:
        k (int):  The number of singular values to be penalized. Default: 1

    Shape:
        - Input: :math:`(b, |\mathcal{f}|)` where :math:`b` is the batch size and :math:`|\mathcal{f}|` is feature dimension.
        - Output: scalar.

    """

    def __init__(self, k: int = 1, weight: float = 1.0):
        super(BatchSpectralShrinkage, self).__init__()
        self.k = k
        self.weight = weight

    def forward(self, feature, **kwargs):
        result = 0
        u, s, v = torch.svd(feature.t())
        num = s.size(0)
        for i in range(self.k):
            result += torch.pow(s[num - 1 - i], 2)
        return self.weight * result



"""
Modified from https://github.com/thuml/Transfer-Learning-Library/blob/master/tllib/normalization/stochnorm.py
"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter

__all__ = ["StochNorm1d", "StochNorm2d", "convert_model"]


class _StochNorm(nn.Module):
    def __init__(
        self, num_features, eps=1e-5, momentum=0.1, affine=True, track_running_stats=True, p=0.5
    ):
        super(_StochNorm, self).__init__()
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.track_running_stats = track_running_stats
        self.p = p
        if self.affine:
            self.weight = Parameter(torch.Tensor(num_features))
            self.bias = Parameter(torch.Tensor(num_features))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

        if self.track_running_stats:
            self.register_buffer("running_mean", torch.zeros(num_features))
            self.register_buffer("running_var", torch.ones(num_features))
        else:
            self.register_parameter("running_mean", None)
            self.register_parameter("running_var", None)
        self.reset_parameters()

    def reset_parameters(self):
        if self.track_running_stats:
            self.running_mean.zero_()
            self.running_var.fill_(1)
        if self.affine:
            self.weight.data.uniform_()
            self.bias.data.zero_()

    def _check_input_dim(self, input):
        return NotImplemented

    def forward(self, input):
        self._check_input_dim(input)

        if self.training:
            z_0 = F.batch_norm(
                input,
                self.running_mean,
                self.running_var,
                self.weight,
                self.bias,
                False,
                self.momentum,
                self.eps,
            )

            z_1 = F.batch_norm(
                input,
                self.running_mean,
                self.running_var,
                self.weight,
                self.bias,
                True,
                self.momentum,
                self.eps,
            )

            if input.dim() == 2:
                s = (
                    torch.from_numpy(
                        np.random.binomial(n=1, p=self.p, size=self.num_features).reshape(  # type: ignore
                            1, self.num_features
                        )
                    )
                    .float()
                    .cuda()
                )
            elif input.dim() == 3:
                s = (
                    torch.from_numpy(
                        np.random.binomial(n=1, p=self.p, size=self.num_features).reshape(  # type: ignore
                            1, self.num_features, 1
                        )
                    )
                    .float()
                    .cuda()
                )
            elif input.dim() == 4:
                s = (
                    torch.from_numpy(
                        np.random.binomial(n=1, p=self.p, size=self.num_features).reshape(  # type: ignore
                            1, self.num_features, 1, 1
                        )
                    )
                    .float()
                    .cuda()
                )
            else:
                raise BaseException()

            z = (1 - s) * z_0 + s * z_1
        else:
            z = F.batch_norm(
                input,
                self.running_mean,
                self.running_var,
                self.weight,
                self.bias,
                False,
                self.momentum,
                self.eps,
            )

        return z


class StochNorm1d(_StochNorm):
    r"""Applies Stochastic Normalization over a 2D or 3D input (a mini-batch of 1D inputs with optional additional channel dimension)

    Stochastic  Normalization is proposed in `Stochastic Normalization (NIPS 2020) <https://papers.nips.cc/paper/2020/file/bc573864331a9e42e4511de6f678aa83-Paper.pdf>`_

    .. math::

        \hat{x}_{i,0} = \frac{x_i - \tilde{\mu}}{ \sqrt{\tilde{\sigma} + \epsilon}}

        \hat{x}_{i,1} = \frac{x_i - \mu}{ \sqrt{\sigma + \epsilon}}

        \hat{x}_i = (1-s)\cdot \hat{x}_{i,0} + s\cdot \hat{x}_{i,1}

         y_i = \gamma \hat{x}_i + \beta

    where :math:`\mu` and :math:`\sigma` are mean and variance of current mini-batch data.

    :math:`\tilde{\mu}` and :math:`\tilde{\sigma}` are current moving statistics of training data.

    :math:`s` is a branch-selection variable generated from a Bernoulli distribution, where :math:`P(s=1)=p`.


    During training, there are two normalization branches. One uses mean and
    variance of current mini-batch data, while the other uses current moving
    statistics of the training data as usual batch normalization.

    During evaluation, the moving statistics is used for normalization.


    Args:
        num_features (int): :math:`c` from an expected input of size :math:`(b, c, l)` or  :math:`l` from an expected input of size :math:`(b, l)`.
        eps (float): A value added to the denominator for numerical stability.
            Default: 1e-5
        momentum (float): The value used for the running_mean and running_var
            computation. Default: 0.1
        affine (bool): A boolean value that when set to ``True``, gives the layer learnable
            affine parameters. Default: ``True``
        track_running_stats (bool): A boolean value that when set to True, this module tracks
         the running mean and variance, and when set to False, this module does not
         track such statistics, and initializes statistics buffers running_mean and
         running_var as None. When these buffers are None, this module always uses
         batch statistics in both training and eval modes. Default: True
         p (float): The probability to choose the second branch (usual BN). Default: 0.5

    Shape:
        - Input: :math:`(b, l)` or :math:`(b, c, l)`
        - Output: :math:`(b, l)` or :math:`(b, c, l)` (same shape as input)
    """

    def _check_input_dim(self, input):
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError("expected 2D or 3D input (got {}D input)".format(input.dim()))


class StochNorm2d(_StochNorm):
    r"""
    Applies Stochastic  Normalization over a 4D input (a mini-batch of 2D inputs with additional channel dimension)

    Stochastic  Normalization is proposed in `Stochastic Normalization (NIPS 2020) <https://papers.nips.cc/paper/2020/file/bc573864331a9e42e4511de6f678aa83-Paper.pdf>`_

    .. math::

        \hat{x}_{i,0} = \frac{x_i - \tilde{\mu}}{ \sqrt{\tilde{\sigma} + \epsilon}}

        \hat{x}_{i,1} = \frac{x_i - \mu}{ \sqrt{\sigma + \epsilon}}

        \hat{x}_i = (1-s)\cdot \hat{x}_{i,0} + s\cdot \hat{x}_{i,1}

         y_i = \gamma \hat{x}_i + \beta

    where :math:`\mu` and :math:`\sigma` are mean and variance of current mini-batch data.

    :math:`\tilde{\mu}` and :math:`\tilde{\sigma}` are current moving statistics of training data.

    :math:`s` is a branch-selection variable generated from a Bernoulli distribution, where :math:`P(s=1)=p`.


    During training, there are two normalization branches. One uses mean and
    variance of current mini-batch data, while the other uses current moving
    statistics of the training data as usual batch normalization.

    During evaluation, the moving statistics is used for normalization.


    Args:
        num_features (int): :math:`c` from an expected input of size :math:`(b, c, h, w)`.
        eps (float): A value added to the denominator for numerical stability.
            Default: 1e-5
        momentum (float): The value used for the running_mean and running_var
            computation. Default: 0.1
        affine (bool): A boolean value that when set to ``True``, gives the layer learnable
            affine parameters. Default: ``True``
        track_running_stats (bool): A boolean value that when set to True, this module tracks
         the running mean and variance, and when set to False, this module does not
         track such statistics, and initializes statistics buffers running_mean and
         running_var as None. When these buffers are None, this module always uses
         batch statistics in both training and eval modes. Default: True
         p (float): The probability to choose the second branch (usual BN). Default: 0.5

    Shape:
        - Input: :math:`(b, c, h, w)`
        - Output: :math:`(b, c, h, w)` (same shape as input)
    """

    def _check_input_dim(self, input):
        if input.dim() != 4:
            raise ValueError("expected 4D input (got {}D input)".format(input.dim()))


class StochNorm3d(_StochNorm):
    r"""
    Applies Stochastic  Normalization over a 5D input (a mini-batch of 3D inputs with additional channel dimension)

    Stochastic  Normalization is proposed in `Stochastic Normalization (NIPS 2020) <https://papers.nips.cc/paper/2020/file/bc573864331a9e42e4511de6f678aa83-Paper.pdf>`_

    .. math::

        \hat{x}_{i,0} = \frac{x_i - \tilde{\mu}}{ \sqrt{\tilde{\sigma} + \epsilon}}

        \hat{x}_{i,1} = \frac{x_i - \mu}{ \sqrt{\sigma + \epsilon}}

        \hat{x}_i = (1-s)\cdot \hat{x}_{i,0} + s\cdot \hat{x}_{i,1}

         y_i = \gamma \hat{x}_i + \beta

    where :math:`\mu` and :math:`\sigma` are mean and variance of current mini-batch data.

    :math:`\tilde{\mu}` and :math:`\tilde{\sigma}` are current moving statistics of training data.

    :math:`s` is a branch-selection variable generated from a Bernoulli distribution, where :math:`P(s=1)=p`.


    During training, there are two normalization branches. One uses mean and
    variance of current mini-batch data, while the other uses current moving
    statistics of the training data as usual batch normalization.

    During evaluation, the moving statistics is used for normalization.


    Args:
        num_features (int): :math:`c` from an expected input of size :math:`(b, c, d, h, w)`
        eps (float): A value added to the denominator for numerical stability.
            Default: 1e-5
        momentum (float): The value used for the running_mean and running_var
            computation. Default: 0.1
        affine (bool): A boolean value that when set to ``True``, gives the layer learnable
            affine parameters. Default: ``True``
        track_running_stats (bool): A boolean value that when set to True, this module tracks
         the running mean and variance, and when set to False, this module does not
         track such statistics, and initializes statistics buffers running_mean and
         running_var as None. When these buffers are None, this module always uses
         batch statistics in both training and eval modes. Default: True
         p (float): The probability to choose the second branch (usual BN). Default: 0.5

    Shape:
        - Input: :math:`(b, c, d, h, w)`
        - Output: :math:`(b, c, d, h, w)` (same shape as input)
    """

    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError("expected 4D input (got {}D input)".format(input.dim()))


def convert_model(module: nn.Module, p: float = 0.5) -> nn.Module:
    """
    Traverses the input module and its child recursively and replaces all
    instance of BatchNorm to StochNorm.

    Args:
        module (torch.nn.Module): The input module needs to be convert to StochNorm model.
        p (float): The hyper-parameter for StochNorm layer.

    Returns:
         The module converted to StochNorm version.
    """

    mod = module
    for pth_module, stoch_module in zip(
        [
            torch.nn.modules.batchnorm.BatchNorm1d,
            torch.nn.modules.batchnorm.BatchNorm2d,
            torch.nn.modules.batchnorm.BatchNorm3d,
        ],
        [StochNorm1d, StochNorm2d, StochNorm3d],
    ):
        if isinstance(module, pth_module):
            mod = stoch_module(module.num_features, module.eps, module.momentum, module.affine, p)  # type: ignore
            mod.running_mean = module.running_mean  # type: ignore
            mod.running_var = module.running_var  # type: ignore

            if module.affine:
                mod.weight.data = module.weight.data.clone().detach()
                mod.bias.data = module.bias.data.clone().detach()

    for name, child in module.named_children():
        mod.add_module(name, convert_model(child, p))

    return mod



"""Transforms Factory
Factory methods for building image transforms for use with TIMM (PyTorch Image Models)

Hacked together by / Copyright 2019, Ross Wightman
"""

from typing import Optional, Tuple, Union

import torch
from torchvision import transforms

from timm.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.data.auto_augment import (
    rand_augment_transform,
    augment_and_mix_transform,
    auto_augment_transform,
)
from timm.data.transforms import (
    str_to_pil_interp,
    RandomResizedCropAndInterpolation,
    ResizeKeepRatio,
    CenterCropOrPad,
    RandomCropOrPad,
    ToNumpy,
    MaybeToTensor,
    MaybePILToTensor,
)
from timm.data.random_erasing import RandomErasing
from timm.data.transforms_factory import (
    transforms_imagenet_eval,
    transforms_noaug_train,
)

def transforms_imagenet_train(
    img_size: Union[int, Tuple[int, int]] = 224,
    scale: Optional[Tuple[float, float]] = None,
    ratio: Optional[Tuple[float, float]] = None,
    train_crop_mode: Optional[str] = None,
    hflip: float = 0.5,
    vflip: float = 0.0,
    color_jitter: Union[float, Tuple[float, ...]] = 0.4,
    color_jitter_prob: Optional[float] = None,
    force_color_jitter: bool = False,
    grayscale_prob: float = 0.0,
    gaussian_blur_prob: float = 0.0,
    auto_augment: Optional[str] = None,
    trivial_augment: bool = False,
    rand_augment: bool = False,
    ra_num_ops: int = 2,
    ra_magnitude: int = 8,
    ra_magnitude_bins: int = 31,
    interpolation: str = "random",
    mean: Tuple[float, ...] = IMAGENET_DEFAULT_MEAN,
    std: Tuple[float, ...] = IMAGENET_DEFAULT_STD,
    re_prob: float = 0.0,
    re_mode: str = "const",
    re_count: int = 1,
    re_num_splits: int = 0,
    use_prefetcher: bool = False,
    normalize: bool = True,
    separate: bool = False,
):
    """ImageNet-oriented image transforms for training.

    Args:
        img_size: Target image size.
        train_crop_mode: Training random crop mode ('rrc', 'rkrc', 'rkrr').
        scale: Random resize scale range (crop area, < 1.0 => zoom in).
        ratio: Random aspect ratio range (crop ratio for RRC, ratio adjustment factor for RKR).
        hflip: Horizontal flip probability.
        vflip: Vertical flip probability.
        color_jitter: Random color jitter component factors (brightness, contrast, saturation, hue).
            Scalar is applied as (scalar,) * 3 (no hue).
        color_jitter_prob: Apply color jitter with this probability if not None (for SimlCLR-like aug).
        force_color_jitter: Force color jitter where it is normally disabled (ie with RandAugment on).
        grayscale_prob: Probability of converting image to grayscale (for SimCLR-like aug).
        gaussian_blur_prob: Probability of applying gaussian blur (for SimCLR-like aug).
        auto_augment: Auto augment configuration string (see auto_augment.py).
        trivial_augment: Add Trivial augment.
        rand_augment: Add RandAugment augmentations.
        ra_num_ops: Number of augmentation transformations in RandAugment.
        ra_magnitude: Magnitude of transformations in RandAugment.
        interpolation: Image interpolation mode.
        mean: Image normalization mean.
        std: Image normalization standard deviation.
        re_prob: Random erasing probability.
        re_mode: Random erasing fill mode.
        re_count: Number of random erasing regions.
        re_num_splits: Control split of random erasing across batch size.
        use_prefetcher: Prefetcher enabled. Do not convert image to tensor or normalize.
        normalize: Normalize tensor output w/ provided mean/std (if prefetcher not used).
        separate: Output transforms in 3-stage tuple.

    Returns:
        If separate==True, the transforms are returned as a tuple of 3 separate transforms
        for use in a mixing dataset that passes
         * all data through the first (primary) transform, called the 'clean' data
         * a portion of the data through the secondary transform
         * normalizes and converts the branches above with the third, final transform
    """
    train_crop_mode = train_crop_mode or "rrc"
    assert train_crop_mode in {"rrc", "rkrc", "rkrr"}
    if train_crop_mode in ("rkrc", "rkrr"):
        # FIXME integration of RKR is a WIP
        scale = tuple(scale or (0.8, 1.00))  # type: ignore
        ratio = tuple(ratio or (0.9, 1 / 0.9))  # type: ignore
        primary_tfl = [
            ResizeKeepRatio(
                img_size,
                interpolation=interpolation,
                random_scale_prob=0.5,
                random_scale_range=scale,
                random_scale_area=True,  # scale compatible with RRC
                random_aspect_prob=0.5,
                random_aspect_range=ratio,
            ),
            CenterCropOrPad(img_size, padding_mode="reflect")  # type: ignore
            if train_crop_mode == "rkrc"
            else RandomCropOrPad(img_size, padding_mode="reflect"),  # type: ignore
        ]
    else:
        scale = tuple(scale or (0.08, 1.0))  # type: ignore
        ratio = tuple(ratio or (3.0 / 4.0, 4.0 / 3.0))  # type: ignore
        primary_tfl = [
            RandomResizedCropAndInterpolation(
                img_size,
                scale=scale,
                ratio=ratio,
                interpolation=interpolation,
            )
        ]
    if hflip > 0.0:
        primary_tfl += [transforms.RandomHorizontalFlip(p=hflip)]
    if vflip > 0.0:
        primary_tfl += [transforms.RandomVerticalFlip(p=vflip)]

    secondary_tfl = []
    disable_color_jitter = False
    if auto_augment:
        assert isinstance(auto_augment, str)
        # color jitter is typically disabled if AA/RA on,
        # this allows override without breaking old hparm cfgs
        disable_color_jitter = not (force_color_jitter or "3a" in auto_augment)
        if isinstance(img_size, (tuple, list)):
            img_size_min = min(img_size)
        else:
            img_size_min = img_size
        aa_params = dict(
            translate_const=int(img_size_min * 0.45),
            img_mean=tuple([min(255, round(255 * x)) for x in mean]),
        )
        if interpolation and interpolation != "random":
            aa_params["interpolation"] = str_to_pil_interp(interpolation)
        if auto_augment.startswith("rand"):
            secondary_tfl += [rand_augment_transform(auto_augment, aa_params)]
        elif auto_augment.startswith("augmix"):
            aa_params["translate_pct"] = 0.3  # type: ignore
            secondary_tfl += [augment_and_mix_transform(auto_augment, aa_params)]
        else:
            secondary_tfl += [auto_augment_transform(auto_augment, aa_params)]

    if color_jitter is not None and not disable_color_jitter:
        # color jitter is enabled when not using AA or when forced
        if isinstance(color_jitter, (list, tuple)):
            # color jitter should be a 3-tuple/list if spec brightness/contrast/saturation
            # or 4 if also augmenting hue
            assert len(color_jitter) in (3, 4)
        else:
            # if it's a scalar, duplicate for brightness, contrast, and saturation, no hue
            color_jitter = (float(color_jitter),) * 3
        if color_jitter_prob is not None:
            secondary_tfl += [
                transforms.RandomApply(
                    [
                        transforms.ColorJitter(*color_jitter),
                    ],
                    p=color_jitter_prob,
                )
            ]
        else:
            secondary_tfl += [transforms.ColorJitter(*color_jitter)]

    if grayscale_prob:
        secondary_tfl += [transforms.RandomGrayscale(p=grayscale_prob)]

    if gaussian_blur_prob:
        secondary_tfl += [
            transforms.RandomApply(
                [
                    transforms.GaussianBlur(kernel_size=23),  # hardcoded for now
                ],
                p=gaussian_blur_prob,
            )
        ]

    if trivial_augment:
        secondary_tfl += [transforms.TrivialAugmentWide()]

    if rand_augment:
        secondary_tfl += [transforms.RandAugment(ra_num_ops, ra_magnitude, ra_magnitude_bins)]

    final_tfl = []
    if use_prefetcher:
        # prefetcher and collate will handle tensor conversion and norm
        final_tfl += [ToNumpy()]
    elif not normalize:
        # when normalize disable, converted to tensor without scaling, keeps original dtype
        final_tfl += [MaybePILToTensor()]
    else:
        final_tfl += [
            MaybeToTensor(),
            transforms.Normalize(
                mean=torch.tensor(mean),
                std=torch.tensor(std),
            ),
        ]
        if re_prob > 0.0:
            final_tfl += [
                RandomErasing(
                    re_prob,
                    mode=re_mode,
                    max_count=re_count,
                    num_splits=re_num_splits,
                    device="cpu",
                )
            ]

    if separate:
        return (
            transforms.Compose(primary_tfl),
            transforms.Compose(secondary_tfl),
            transforms.Compose(final_tfl),
        )
    else:
        return transforms.Compose(primary_tfl + secondary_tfl + final_tfl)


def create_transform(
    input_size: Union[int, Tuple[int, int], Tuple[int, int, int]] = 224,
    is_training: bool = False,
    no_aug: bool = False,
    train_crop_mode: Optional[str] = None,
    scale: Optional[Tuple[float, float]] = None,
    ratio: Optional[Tuple[float, float]] = None,
    hflip: float = 0.5,
    vflip: float = 0.0,
    color_jitter: Union[float, Tuple[float, ...]] = 0.4,
    color_jitter_prob: Optional[float] = None,
    grayscale_prob: float = 0.0,
    gaussian_blur_prob: float = 0.0,
    auto_augment: Optional[str] = None,
    trivial_augment: bool = False,
    rand_augment: bool = False,
    ra_num_ops: int = 2,
    ra_magnitude: int = 9,
    ra_magnitude_bins: int = 31,
    interpolation: str = "bilinear",
    mean: Tuple[float, ...] = IMAGENET_DEFAULT_MEAN,
    std: Tuple[float, ...] = IMAGENET_DEFAULT_STD,
    re_prob: float = 0.0,
    re_mode: str = "const",
    re_count: int = 1,
    re_num_splits: int = 0,
    crop_pct: Optional[float] = None,
    crop_mode: Optional[str] = None,
    crop_border_pixels: Optional[int] = None,
    tf_preprocessing: bool = False,
    use_prefetcher: bool = False,
    normalize: bool = True,
    separate: bool = False,
):
    """

    Args:
        input_size: Target input size (channels, height, width) tuple or size scalar.
        is_training: Return training (random) transforms.
        no_aug: Disable augmentation for training (useful for debug).
        train_crop_mode: Training random crop mode ('rrc', 'rkrc', 'rkrr').
        scale: Random resize scale range (crop area, < 1.0 => zoom in).
        ratio: Random aspect ratio range (crop ratio for RRC, ratio adjustment factor for RKR).
        hflip: Horizontal flip probability.
        vflip: Vertical flip probability.
        color_jitter: Random color jitter component factors (brightness, contrast, saturation, hue).
            Scalar is applied as (scalar,) * 3 (no hue).
        color_jitter_prob: Apply color jitter with this probability if not None (for SimlCLR-like aug).
        grayscale_prob: Probability of converting image to grayscale (for SimCLR-like aug).
        gaussian_blur_prob: Probability of applying gaussian blur (for SimCLR-like aug).
        auto_augment: Auto augment configuration string (see auto_augment.py).
        interpolation: Image interpolation mode.
        mean: Image normalization mean.
        std: Image normalization standard deviation.
        re_prob: Random erasing probability.
        re_mode: Random erasing fill mode.
        re_count: Number of random erasing regions.
        re_num_splits: Control split of random erasing across batch size.
        crop_pct: Inference crop percentage (output size / resize size).
        crop_mode: Inference crop mode. One of ['squash', 'border', 'center']. Defaults to 'center' when None.
        crop_border_pixels: Inference crop border of specified # pixels around edge of original image.
        tf_preprocessing: Use TF 1.0 inference preprocessing for testing model ports
        use_prefetcher: Pre-fetcher enabled. Do not convert image to tensor or normalize.
        normalize: Normalization tensor output w/ provided mean/std (if prefetcher not used).
        separate: Output transforms in 3-stage tuple.

    Returns:
        Composed transforms or tuple thereof
    """
    if isinstance(input_size, (tuple, list)):
        img_size = input_size[-2:]
    else:
        img_size = input_size

    if tf_preprocessing and use_prefetcher:
        assert not separate, "Separate transforms not supported for TF preprocessing"
        from timm.data.tf_preprocessing import TfPreprocessTransform

        transform = TfPreprocessTransform(
            is_training=is_training,
            size=img_size,  # type: ignore
            interpolation=interpolation,
        )
    else:
        if is_training and no_aug:
            assert not separate, "Cannot perform split augmentation with no_aug"
            transform = transforms_noaug_train(
                img_size,
                interpolation=interpolation,
                mean=mean,
                std=std,
                use_prefetcher=use_prefetcher,
                normalize=normalize,
            )
        elif is_training:
            transform = transforms_imagenet_train(
                img_size,
                train_crop_mode=train_crop_mode,
                scale=scale,
                ratio=ratio,
                hflip=hflip,
                vflip=vflip,
                color_jitter=color_jitter,
                color_jitter_prob=color_jitter_prob,
                grayscale_prob=grayscale_prob,
                gaussian_blur_prob=gaussian_blur_prob,
                auto_augment=auto_augment,
                trivial_augment=trivial_augment,
                rand_augment=rand_augment,
                ra_num_ops=ra_num_ops,
                ra_magnitude=ra_magnitude,
                ra_magnitude_bins=ra_magnitude_bins,
                interpolation=interpolation,
                mean=mean,
                std=std,
                re_prob=re_prob,
                re_mode=re_mode,
                re_count=re_count,
                re_num_splits=re_num_splits,
                use_prefetcher=use_prefetcher,
                normalize=normalize,
                separate=separate,
            )
        else:
            assert not separate, "Separate transforms not supported for validation preprocessing"
            transform = transforms_imagenet_eval(
                img_size,
                interpolation=interpolation,
                mean=mean,
                std=std,
                crop_pct=crop_pct,
                crop_mode=crop_mode,
                crop_border_pixels=crop_border_pixels,
                use_prefetcher=use_prefetcher,
                normalize=normalize,
            )

    return transform
